#include "Audio.h"
#include <FK/Error.h>
#include <FK/Model.h>

#include <iostream>
#include <vector>

#ifdef __BIG_ENDIAN__
const int _ENDIAN = 1;
#else
const int _ENDIAN = 0;
#endif

using namespace std;

static fk_IDAdmin * getAdmin(void)
{
	static fk_IDAdmin audioIDAdmin(0);
	return &audioIDAdmin;
}

static bool					initStatus = false;
static ALCcontext			*alContext = NULL;

fk_Model *fk_AudioBase::listenerCamera = NULL;


static bool ALInit(void)
{
	ALCdevice	*device;
	ALCcontext	*context;

	device = alcOpenDevice(NULL);
	if(device == NULL) {
		return false;
	}

	context = alcCreateContext(device, NULL);
	if(context == NULL) {
		alcCloseDevice (device);
		return false;
    }

	if(!alcMakeContextCurrent(context)) {
		alcDestroyContext(context);
		alcCloseDevice(device);
		return false;
	}

	alContext = context;
	return true;
}

static bool ALExit(void)
{
	ALCdevice *device;

	if(!alcMakeContextCurrent(NULL)) {
		return false;
	}

	device = alcGetContextsDevice(alContext);
	alcDestroyContext(alContext);
	if(alcGetError(device) != ALC_NO_ERROR) {
		return false;
	}

	if(!alcCloseDevice(device)) {
		return false;
	}

	alContext = NULL;
	return true;
}

#if HAVE_NANOSLEEP && HAVE_TIME_H
#include <time.h>
#include <errno.h>
#elif HAVE_USLEEP && HAVE_UNISTD_H
#include <unistd.h>
#elif WIN32
#include <windows.h>
#else
#include <time.h>
#include <errno.h>
//#error No way to sleep on this platform
#endif

bool ALSleep (ALfloat duration)
{
	if(duration < 0) {
		return false;
	}

	{
		ALuint seconds = (ALuint) duration;
		ALfloat rest = duration - (ALfloat) seconds;

#if HAVE_NANOSLEEP && HAVE_TIME_H

		ALuint microSecs = (ALuint) (rest * 1000000);
		struct timespec t, remainingTime;
		t.tv_sec = (time_t) seconds;
		t.tv_nsec = ((long) microSecs) * 1000;

		/* At least the interaction of nanosleep and signals is specified! */
		while(nanosleep (&t, &remainingTime) < 0) {
			if(errno != EINTR) {
				return false;
			}
			/* If we received a signal, let's try again with the remaining time. */
			t.tv_sec = remainingTime.tv_sec;
			t.tv_nsec = remainingTime.tv_nsec;
		}

#elif HAVE_USLEEP && HAVE_UNISTD_H

		while (seconds > 0) {
			usleep (1000000);
			seconds--;
		}
		usleep ((unsigned int) (rest * 1000000));

#elif WIN32

		while(seconds > 0) {
			Sleep(1000);
			seconds--;
		}
		Sleep((DWORD)(rest*1000));

#endif

	}
	return true;
}


fk_AudioBase::fk_AudioBase(void)
{
	queueSize = FK_OV_DEFAULT_QUEUE_SIZE;
	format = AL_FORMAT_STEREO16;
	rate = 0;
	source_id = -1;
	source = 99999;
	gain = 1.0;
	loopMode = false;
	loopStartTime = 0.0;
	loopEndTime = -1.0;
	init();

	return;
}

fk_AudioBase::~fk_AudioBase()
{
	if(source_id != -1) idAdmin.EraseID(source_id);
	if(idAdmin.GetIDNum() != 0) return;

	if(initStatus) {
		if(!ALExit()) {
			cerr << "ALExit() error." << endl;
		}
		initStatus = false;
	}
}

bool fk_AudioBase::init(void)
{
	if(initStatus == false) {
		if(!ALInit()) {
			cerr << "ALInit() error." << endl;
			return false;
		}
	}
	initStatus = true;

	return true;
}

bool fk_AudioBase::getInit(void)
{
	return initStatus;
}

void fk_AudioBase::CreateID(void)
{
	source_id = idAdmin.CreateID();
	alGenSources(1, &source);
	alSourcef(source, AL_GAIN, ALfloat(gain));
	alSourcei(source, AL_SOURCE_RELATIVE, AL_TRUE);
	return;
}

void fk_AudioBase::EraseID(void)
{
	idAdmin.EraseID(source_id);
	source_id = -1;
	return;
}

void fk_AudioBase::setGain(double argGain)
{
	if(argGain < -0.00001 || argGain > 1.00001) return;
	gain = argGain;
	if(source_id != -1) {
		alSourcef(source, AL_GAIN, ALfloat(gain));
	}
	return;
}

double fk_AudioBase::getGain(void)
{
	return gain;
}


void fk_AudioBase::setQueueSize(int argSize)
{
	if(argSize < 1) return;
	queueSize = argSize;
	return;
}

int fk_AudioBase::getQueueSize(void)
{
	return queueSize;
}

void fk_AudioBase::setLoopMode(bool argMode)
{
	loopMode = argMode;
	return;
}

bool fk_AudioBase::getLoopMode(void)
{
	return loopMode;
}

void fk_AudioBase::setLoopArea(double argST, double argED)
{
	loopStartTime = argST;
	loopEndTime = argED;
	return;
}

double fk_AudioBase::getLoopStartTime(void)
{
	return loopStartTime;
}

double fk_AudioBase::getLoopEndTime(void)
{
	return loopEndTime;
}

void fk_AudioBase::MakeOVInfo(OggVorbis_File *argVF)
{
	vorbis_info		*info = NULL;

	info = ov_info(argVF, -1);
	if(info->channels == 1) {
		format = AL_FORMAT_MONO16;
	} else {
		format = AL_FORMAT_STEREO16;
	}
	rate = info->rate;

	return;
}

void fk_AudioBase::sleep(double argTime)
{
	ALSleep(float(argTime));
	return;
}

void fk_AudioBase::pause(void)
{
	ALint		status;

	if(source_id == -1) return;
	alGetSourcei(source, AL_SOURCE_STATE, &status);
	if(status == AL_PLAYING) {
		alSourcePause(source);
	}
	return;
}

void fk_AudioBase::setListenerModel(fk_Model *argModel)
{
	listenerCamera = argModel;
}

void fk_AudioBase::setListenerModel(fk_Model &argModel)
{
	listenerCamera = &argModel;
}

fk_Model * fk_AudioBase::getListenerModel(void)
{
	return listenerCamera;
}

void fk_AudioBase::UpdateListener(void)
{
	static fk_Vector	camPos, camVec, camUpvec;
	static ALfloat		orient[6];

	if(listenerCamera != NULL) {
		camPos = listenerCamera->getInhPosition();
		camVec = listenerCamera->getInhVec();
		camUpvec = listenerCamera->getInhUpvec();
		orient[0] = (float)camVec.x;	orient[1] = (float)camVec.y;	orient[2] = (float)camVec.z;
		orient[3] = (float)camUpvec.x;	orient[4] = (float)camUpvec.y;	orient[5] = (float)camUpvec.z;
		alListener3f(AL_POSITION, (float)camPos.x, (float)camPos.y, (float)camPos.z); 
		alListenerfv(AL_ORIENTATION, orient); 
	}

	return;
}
